package ch.sahits.game.tetris.ctrl;

import java.awt.Color;
import java.awt.Point;
import java.awt.event.KeyEvent;

import ch.sahits.game.event.Event;
import ch.sahits.game.event.IEventListener;
import ch.sahits.game.tetris.event.*;
import ch.sahits.game.tetris.model.Block;
import ch.sahits.game.tetris.model.BuildingBlock;
import ch.sahits.game.tetris.model.EBuildingBlocks;
import ch.sahits.game.tetris.model.KeyMap;
import ch.sahits.game.tetris.model.PlayingField;
import ch.sahits.game.tetris.model.TetrisViewModel;
import ch.sahits.game.tetris.ui.TetrisConstants;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.*;
/**
 * Game controller
 * @author andi
 *
 */
public class TetrisController implements IEventListener{

	private Level level;
	private int uncoveredLines = 0;
	private int nextUpdate = -1;
	private Randomizer rnd = new Randomizer();
	private EBuildingBlocks nextShape;
	private EGameState currentState = EGameState.START;
	
	private TetrisViewModel viewModel;
	
	public TetrisController() {
		Event.add(this);
	}
	
	public void gameUpdate() {
		if (!isPlayState(currentState) || nextUpdate == -1){
			return;
		}
		if (nextUpdate == 0 && isPlayState(currentState)) {
			moveOneDown();
		} else {
			nextUpdate--;
		}
	}
	/**
	 * Move the block down by one.
	 */
	private void moveOneDown() {
		BuildingBlock bb = viewModel.getBlock().toBuildingBlock();
		if (bb.canMoveDown(viewModel.getPlayingField())){
			Block newBlock = bb.moveDown();
			viewModel.setBlock(newBlock);
			new PlayFieldUpdateEvent(viewModel).notify(null);
		} else {
			updateViewModelWithBlock();
			// check remove line
			new PlayFieldUpdateEvent(viewModel).notify(null);
			nextShape = rnd.nextShape();
			new NextShapeEvent().notify(nextShape);
		}
		nextUpdate = level.getSpeed();
	}
	/**
	 * add the block to the fields, create a new block and
	 * check if the row was coompleted. Update the model.
	 */
	private void updateViewModelWithBlock() {
		Point[] positions = viewModel.getBlock().getPosition();
		PlayingField field = viewModel.getPlayingField();
		int minRow = TetrisConstants.NUMBER_OF_ROWS;
		int maxRow = 0;
		for (Point point : positions) {
			if (point.y > maxRow) {
				maxRow = point.y;
			}
			if (point.y < minRow) {
				minRow = point.y;
			}
			field.set(point.x, point.y, viewModel.getBlock().getColor());
		}
		for ( int i = minRow; i<=maxRow; i++){
			Color[] row = field.getRow(i);
			boolean rowComplete = true;
			for (Color color : row) {
				if (color == null && rowComplete) {
					rowComplete = false;
				}
			}
			if (rowComplete){
				new RowCompleteEvent().notify(i);
				// remove row
				for (int x=0;x<TetrisConstants.BLOCKS_PER_ROW;x++){
					field.set(x, i, null);
				}
				for (int y=i; y>0;y--){
					for (int x=0;x<TetrisConstants.BLOCKS_PER_ROW;x++){
						Color color = field.get(x, y-1);
						field.set(x, y, color);
					}
				}
				for (int x=0;x<TetrisConstants.BLOCKS_PER_ROW;x++){
					field.set(x, 0, null);
				}
				new PlayFieldUpdateEvent(viewModel).notify(null);
			}
		}
		Block block = new Block(nextShape, initialPosition(nextShape), findCenter(nextShape));
		for (Point point : block.getPosition()){
			if (field.get(point.x, point.y)!=null){
				// game over
				new GameStateChangeEvent(EGameState.GAME_OVER).notify(null);
				return;
			}
		}
		
		nextShape = rnd.nextShape();
		viewModel.setBlock(block);
	}

	public void gameUpdate(Event e, Object notice) {
		if (e instanceof RowCompleteEvent) {
			uncoveredLines++;
			if (level.getLines()==uncoveredLines){
				level = level.nextLevel();
				new GameStateChangeEvent(EGameState.LEVEL_COMPLETED).notify(null);
			}
		}
		if (e instanceof GameStateChangeEvent) {
			GameStateChangeEvent change = (GameStateChangeEvent) e;
			if (currentState == EGameState.MENU && change.getNextGameState() == EGameState.PLAY) {
				clearPlayingField(viewModel.getPlayingField());
				new PlayFieldUpdateEvent(viewModel).notify(null);
			}
			currentState = change.getNextGameState();
		}
		if (e instanceof KeyStrokeEvent){
			
			if (GameStateUtility.isKeySetState(currentState) && notice instanceof KeyEvent) {
				KeyEvent keyEvent = (KeyEvent) notice;
				if (currentState == EGameState.SET_KEY_LEFT) {
					KeyMap.LEFT = keyEvent.getKeyCode();
					new GameStateChangeEvent(EGameState.SET_KEY_RIGHT).notify(null);
				} else 
				if (currentState == EGameState.SET_KEY_RIGHT) {
					KeyMap.RIGHT = keyEvent.getKeyCode();
					new GameStateChangeEvent(EGameState.SET_KEY_DOWN).notify(null);
				} else
				if (currentState == EGameState.SET_KEY_DOWN) {
					KeyMap.DOWN = keyEvent.getKeyCode();
					new GameStateChangeEvent(EGameState.SET_KEY_ROTATE).notify(null);
				} else
				if (currentState == EGameState.SET_KEY_ROTATE) {
					KeyMap.ROTATE = keyEvent.getKeyCode();
					new GameStateChangeEvent(EGameState.SET_KEY_PAUSE).notify(null);
				} else
				if (currentState == EGameState.SET_KEY_PAUSE) {
					KeyMap.PAUSE = keyEvent.getKeyCode();
					new GameStateChangeEvent(EGameState.MENU).notify(null);
				}
			}
			
			// start next level
			if (isStateLevelCompleted(currentState) && notice == EKeyStroke.ENTER) {
				new NextLevelEvent().notify(level);
				new GameStateChangeEvent(EGameState.NEXT_LEVEL_INFO).notify(null);
			} else
			if (isStateNextLevelInfo(currentState) && notice == EKeyStroke.ENTER) {
				clearPlayingField(viewModel.getPlayingField());
				new PlayFieldUpdateEvent(viewModel).notify(null);
				new GameStateChangeEvent(EGameState.PLAY).notify(null);
			}
			if (isGameOverState(currentState) && notice == EKeyStroke.ENTER){
				new GameStateChangeEvent(EGameState.MENU).notify(null);
			}
			if (notice == EKeyStroke.PAUSE && (isPlayState(currentState) || isPauseState(currentState))) {
				currentState = togglePause(currentState);
			}
			if (notice == EKeyStroke.DOWN &&  isPlayState(currentState)){
				moveOneDown();
			}
			if (notice == EKeyStroke.LEFT && isPlayState(currentState)){
				BuildingBlock bb = viewModel.getBlock().toBuildingBlock();
				if (bb.canMoveLeft(viewModel.getPlayingField())){
					viewModel.setBlock(bb.moveLeft());
					new PlayFieldUpdateEvent(viewModel).notify(null);
				}
			}
			if (notice == EKeyStroke.RIGHT && isPlayState(currentState)){ 
				BuildingBlock bb = viewModel.getBlock().toBuildingBlock();
				if (bb.canMoveRight(viewModel.getPlayingField())){
					viewModel.setBlock(bb.moveRight());
					new PlayFieldUpdateEvent(viewModel).notify(null);
				}
			}
			if (notice == EKeyStroke.ROTATE && isPlayState(currentState)){
				BuildingBlock bb = viewModel.getBlock().toBuildingBlock();
				if (bb.canRotate(viewModel.getPlayingField())){
					viewModel.setBlock(bb.rotate());
					new PlayFieldUpdateEvent(viewModel).notify(null);
				}
			}
		}
	}

	/**
	 * Initialize the model.
	 * @return
	 */
	public TetrisViewModel initializeViewModel(){
		PlayingField field = new PlayingField();
		clearPlayingField(field);
		EBuildingBlocks type = rnd.nextShape();
		Block block = new Block(type, initialPosition(type), findCenter(type));
		viewModel = new TetrisViewModel();
		viewModel.setBlock(block);
		viewModel.setPlayingField(field);
		nextShape = rnd.nextShape();
		new NextShapeEvent().notify(nextShape);
		level = new Level(10, 25,1); // todo reset to 10
		nextUpdate = level.getSpeed();
		new NextLevelEvent().notify(level);
		return viewModel;
	}
	/**
	 * Clear the playing field.
	 * @param field
	 */
	private void clearPlayingField(PlayingField field) {
		for (int y=0;y<field.getNbRows();y++){
			for (int x=0;x<field.getNbColumns();x++){
				field.set(x, y, null);
			}
		}
	}
	/**
	 * Retrieve the center points of the blocks
	 * @param type
	 * @return
	 */
	private Point findCenter(EBuildingBlocks type) {
		switch (type) {
		case CUBE:
			return new Point(4,1);
		case I:
			return new Point(4,2);
		case L_LEFT:
			return new Point(5,1);
		case L_RIGHT:
			return new Point(4,1);
		case LEFT:
			return new Point(4,1);
		case RIGHT:
			return new Point(5,1);
		case T:
			return new Point(5,1);
		default:
			throw new IllegalStateException("Unknown type: "+type);
		}
	}
	/**
	 * Initialize the positions of the blocks
	 * @param type
	 * @return
	 */
	private Point[] initialPosition(EBuildingBlocks type){
		switch (type) {
		case CUBE:
			return new Point[]{new Point(4,0),new Point(5,0),new Point(4,1),new Point(5,1)};
		case I:
			return new Point[]{new Point(4,0),new Point(4,1),new Point(4,2),new Point(4,3)};
		case L_LEFT:
			return new Point[]{new Point(5,0),new Point(5,1),new Point(4,2),new Point(5,2)};
		case L_RIGHT:
			return new Point[]{new Point(4,0),new Point(4,1),new Point(4,2),new Point(5,2)};
		case LEFT:
			return new Point[]{new Point(5,0),new Point(4,1),new Point(5,1),new Point(4,2)};
		case RIGHT:
			return new Point[]{new Point(4,0),new Point(4,1),new Point(5,1),new Point(5,2)};
		case T:
			return new Point[]{new Point(5,0),new Point(4,1),new Point(5,1),new Point(6,1)};
		default:
			throw new IllegalStateException("Unknown type: "+type);
		}
	}
}
