package ch.sahits.game.tetris.model;

import java.awt.Point;

import ch.sahits.game.tetris.ui.TetrisConstants;

/**
 * Utility class for moving blocks.
 * @author andi
 *
 */
public class BuildingBlock {
	
	private final Block block;
		
	public BuildingBlock(Block block) {
		this.block = block;
		
	}
	/**
	 * Move the bloc down by one.<br>
	 * PRE: canMoveDown == true
	 * @param field
	 * @return
	 */
	public Block moveDown(){
		Point[] positions = block.getPosition();
		for (Point point : positions) {
			point.y++;
		}
		Point center = block.getCenter();
		center.y++;
		block.setCenter(center);
		return block;
	}
	/**
	 * Rotate the block clockwise by 90 degree.<br>
	 * PRE: canRotate == true
	 * @return
	 */
	public Block rotate(){
		Point[] origanl = block.getPosition();
		Point[] newPos = new Point[origanl.length];
		Point center = block.getCenter();
		// Move center -> (0,0)
		for (int i = 0; i < origanl.length; i++) {
			Point p = origanl[i];
			// (x,y) -> (y,-x)
			newPos[i] = move(p, new Point(-center.x,-center.y));
			newPos[i] = new Point(newPos[i].y,-newPos[i].x);
			newPos[i] = move(newPos[i], new Point(center.x,center.y));
		}
		block.setPosition(newPos);
		return block;
	}
	/**
	 * Move a point by
	 * @param p
	 * @param by
	 * @return
	 */
	private Point move(Point p, Point by) {
		return new Point(p.x+by.x,p.y+by.y);
	}
	/**
	 * chack if the block can move down in the playing field
	 * @param field
	 * @return
	 */
	public boolean canMoveDown(PlayingField field){
		for (Point p : block.getPosition()) {
			if (p.y==field.getNbRows()-1){
				return false;
			}
			if (field.get(p.x, p.y+1)!=null){
				return false;
			}
		}
		return true;
	}
	/**
	 * check if the block can move left in the playing field
	 * @param field
	 * @return
	 */
	public boolean canMoveLeft(PlayingField field){
		for (Point p : block.getPosition()) {
			if (p.x>0){
				if (field.get(p.x-1, p.y)!=null){
					return false;
				}
			} else {
				return false;
			}
		}
		return true;
	}
	/**
	 * check if the block can move to the right in the playing field
	 * @param field
	 * @return
	 */
	public boolean canMoveRight(PlayingField field){
		for (Point p : block.getPosition()) {
			if (p.x<TetrisConstants.BLOCKS_PER_ROW-1){
				if (field.get(p.x+1, p.y)!=null){
					return false;
				}
			} else {
				return false;
			}
		}
		return true;
	}
	/**
	 * Move the block to the left by one.<br>
	 * PRE: canMoveLeft == true
	 * @return
	 */
	public Block moveLeft(){
		Point[] positions = block.getPosition();
		for (Point point : positions) {
			point.x--;
		}
		Point center = block.getCenter();
		center.x--;
		block.setCenter(center);
		return block;
	}
	/**
	 * Move the block to the right by one.<br>
	 * PRE: canMoveRight == true
	 * @return
	 */
	public Block moveRight(){
		Point[] positions = block.getPosition();
		for (Point point : positions) {
			point.x++;
		}
		Point center = block.getCenter();
		center.x++;
		block.setCenter(center);
		return block;
	}
	/**
	 * Check if the block can rotate clockwise 90 degries in the playing field
	 * @param field
	 * @return
	 */
	public boolean canRotate(PlayingField field){
		Point center = block.getCenter();
		// check rotating the fields below to the left
		int nbFieldsBelow = computeFieldsBelow(center);
		if (center.x <= nbFieldsBelow) {
			return false;
		}
		for (int y = center.y; y <= center.y + nbFieldsBelow; y++){
			for (int x = center.x - nbFieldsBelow; x <= center.x; x++) {
				if (field.get(x, y) != null) {
					return false;
				}
			}
		}
		
		// check rotating the fields to the left to above
		int nbFieldsLeft = computeFieldsLeft(center);
		for (int y = center.y - nbFieldsLeft; y<=center.y;y++){
			for (int x=center.x-nbFieldsLeft; x<=center.x;x++){
				if (field.get(x, y) != null) {
					return false;
				}
			}
		}
		
		// check rotating above to the right
		int nbFieldsAbove = computeFieldsAbove(center);
		if (center.x >= TetrisConstants.BLOCKS_PER_ROW-nbFieldsAbove){
			return false;
		}
		for (int y=center.y-nbFieldsAbove;y<=center.y;y++){
			for (int x=center.x;x<=center.x+nbFieldsAbove;x++){
				if (field.get(x, y) != null) {
					return false;
				}
			}
		}
		
		// check rotating from the right to below
		int nbFieldsRight = computeFieldsRight(center);
		for (int y=center.y;y<=center.y+nbFieldsRight;y++){
			for (int x=center.x;x<=center.x+nbFieldsRight;x++){
				if (field.get(x, y) != null) {
					return false;
				}
			}
		}
		return true;
	}
	/**
	 * Count the fields of the block to the right of center
	 * @param center
	 * @return
	 */
	private int computeFieldsRight(Point center) {
		int max = center.x;
		for (Point p : block.getPosition()) {
			if (p.x > max) {
				max = p.x;
			}
		}
		return max - center.x;
	}
	/**
	 * Count the fields of the block above the center
	 * @param center
	 * @return
	 */
	private int computeFieldsAbove(Point center) {
		int min = center.y;
		for (Point p : block.getPosition()) {
			if (p.y < min) {
				min = p.y;
			}
		}
		return center.y - min;
	}
	/**
	 * Count the fields of the block to the left of center
	 * @param center
	 * @return
	 */
	private int computeFieldsLeft(Point center) {
		int min = center.x;
		for (Point p : block.getPosition()) {
			if (p.x < min) {
				min = p.x;
			}
		}
		return center.x - min;
	}
	/**
	 * Count the fields of the block below the center
	 * @param center
	 * @return
	 */
	private int computeFieldsBelow(Point center) {
		int max = center.y;
		for (Point p : block.getPosition()) {
			if (p.y > max) {
				max = p.y;
			}
		}
		return max - center.y;
	}

}
