package ch.sahits.game.tetris.ui;

import static ch.sahits.game.tetris.ctrl.GameStateUtility.isMenuState;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.isPauseState;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.isPlayState;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.isStateLevelCompleted;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.isStateNextLevelInfo;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.togglePause;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.font.GlyphVector;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;

import javax.imageio.ImageIO;

import org.apache.commons.lang.WordUtils;

import ch.sahits.game.event.Event;
import ch.sahits.game.image.EScaling;
import ch.sahits.game.image.ImageScaleState;
import ch.sahits.game.image.ImageUtilV2;
import ch.sahits.game.image.Scaling;
import ch.sahits.game.rendering.AbstractRenderPart;
import ch.sahits.game.tetris.ctrl.GameStateUtility;
import ch.sahits.game.tetris.ctrl.Level;
import ch.sahits.game.tetris.event.EGameState;
import ch.sahits.game.tetris.event.EKeyStroke;
import ch.sahits.game.tetris.event.GameStateChangeEvent;
import ch.sahits.game.tetris.event.KeyStrokeEvent;
import ch.sahits.game.tetris.event.NextLevelEvent;
import ch.sahits.game.tetris.event.RowCompleteEvent;
import static ch.sahits.game.tetris.ui.RenderingUtilities.*;
/**
 * The Left side view to show the image and additional informations.
 * @author andi
 *
 */
public class PictureView extends AbstractRenderPart {
	
	private final BufferedImage[] images;
	private final String[] fileNames;
	private Level level;
	private Level nextLevel = null;
	private int showLines = 0;
	private EGameState currentState = EGameState.START;

	public PictureView(Rectangle rect) throws IOException, URISyntaxException {
		super(rect);
		File dir;
		URL root = getClass().getResource("/");
		if (root != null) {
			dir = new File(root.toURI());
		} else {
			dir = new File(".");
		}
		File[] files = dir.listFiles(new FilenameFilter());
		images = new BufferedImage[files.length];
		fileNames = new String[files.length];
		for (int i = 0; i < files.length; i++) {
			images[i] = ImageIO.read(files[i]);
			fileNames[i] = files[i].getName();
		}
		scaleImages(rect);
		Event.add(this);
	}
	/**
	 * Scale the images, so they can be displayed if they are larger
	 * @param rect
	 */
	private void scaleImages(Rectangle rect) {
		ImageUtilV2 imageUtils = new ImageUtilV2();
		Dimension targetDim = new Dimension(rect.width-2*TetrisConstants.PICTURE_VIEW_BORDER, 
				                            rect.height-2*TetrisConstants.PICTURE_VIEW_BORDER);
		for (int i = 0; i < images.length; i++) {
			Dimension origDim = new Dimension(images[i].getWidth(), images[i].getHeight());
			ImageScaleState state = new ImageScaleState(origDim, targetDim);
			Scaling scaling = imageUtils.computeScaleFactor(targetDim, origDim, EScaling.DOWN);
			state.setScaleFactor(scaling.getScaleFactor());
			images[i] = imageUtils.scaleAndBorder(images[i], state);
		}
		
	}
	public void gameRender(Graphics g) {
		Color oldColor = g.getColor();
		renderGameMenu(g);
		renderKeyConfiguration(g);
		renderGamePlay(g);
		g.setColor(oldColor);
	}
	/**
	 * Rendering for the configuration of the keys.
	 * @param g
	 */
	private void renderKeyConfiguration(Graphics g) {
		if (GameStateUtility.isKeySetState(currentState)){
			Color oldColor = g.getColor();
			g.setColor(Color.BLACK);
			g.fillRect(getBounds().x, getBounds().y, getBounds().width, getBounds().height);
			g.setColor(Color.WHITE);
			Graphics2D g2D = (Graphics2D) g;
			Font font = g.getFont();
			font = font.deriveFont(36);
			font = font.deriveFont(Font.BOLD);
			int x = getBounds().x+TetrisConstants.PICTURE_VIEW_BORDER;
			int y = getBounds().y+getBounds().height/2;
			int h;
			GlyphVector gv = createGlyphVector(g2D, "Left .....", font);
			h = gv.getVisualBounds().getBounds().height;
			g2D.drawGlyphVector(gv, x, y);
			if (currentState == EGameState.SET_KEY_RIGHT
					|| currentState == EGameState.SET_KEY_DOWN
					|| currentState == EGameState.SET_KEY_ROTATE
					|| currentState == EGameState.SET_KEY_PAUSE) {
				gv = createGlyphVector(g2D, "Right ....", font);
				y += 2*h;
				h = gv.getVisualBounds().getBounds().height;
				g2D.drawGlyphVector(gv, x, y);
				if (currentState == EGameState.SET_KEY_DOWN
						|| currentState == EGameState.SET_KEY_ROTATE
						|| currentState == EGameState.SET_KEY_PAUSE) {
					gv = createGlyphVector(g2D, "Down .....", font);
					y += 2*h;
					h = gv.getVisualBounds().getBounds().height;
					g2D.drawGlyphVector(gv, x, y);
					if (currentState == EGameState.SET_KEY_ROTATE
							|| currentState == EGameState.SET_KEY_PAUSE) {
						gv = createGlyphVector(g2D, "Rotate ...", font);
						y += 2*h;
						h = gv.getVisualBounds().getBounds().height;
						g2D.drawGlyphVector(gv, x, y);
						if (currentState == EGameState.SET_KEY_PAUSE){
							gv = createGlyphVector(g2D, "Pause ....", font);
							y += 2*h;
							g2D.drawGlyphVector(gv, x, y);
						}
					}
				}
			}
			g.setColor(oldColor);
		}
	}
	/**
	 * Render the menu
	 * @param g
	 */
	private void renderGameMenu(Graphics g) {
		if (isMenuState(currentState)){
			Color oldColor = g.getColor();
			g.setColor(Color.BLACK);
			g.fillRect(getBounds().x, getBounds().y, getBounds().width, getBounds().height);
			g.setColor(Color.WHITE);
			Graphics2D g2D = (Graphics2D) g;
			Font font = g.getFont();
			font = font.deriveFont(36);
			font = font.deriveFont(Font.BOLD);
			GlyphVector gv = createGlyphVector(g2D, "1 Define keys", font);
			int h = gv.getVisualBounds().getBounds().height;
			int x = getBounds().x+TetrisConstants.PICTURE_VIEW_BORDER;
			int y = getBounds().y+getBounds().height/2;
			g2D.drawGlyphVector(gv, x, y);
			gv = createGlyphVector(g2D, "2 Play", font);
			y += 2*h;
			g2D.drawGlyphVector(gv, x, y);
			gv = createGlyphVector(g2D, "3 Exit", font);
			y += 2*h;
			g2D.drawGlyphVector(gv, x, y);		
			g.setColor(oldColor);
		}
	}
	/**
	 * Rendering in game play display
	 * @param g
	 */
	private void renderGamePlay(Graphics g) {
		Color oldColor = g.getColor();
		if (isPlayState(currentState) || isStateLevelCompleted(currentState)){
			g.setColor(Color.BLACK);
			g.fillRect(getBounds().x, getBounds().y, getBounds().width, getBounds().height);
			if (level !=null) {
				BufferedImage image = images[computeImageIndex()];
				int x = getBounds().x+TetrisConstants.PICTURE_VIEW_BORDER;
				int y = getBounds().y+TetrisConstants.PICTURE_VIEW_BORDER;
				int width = image.getWidth();
				int height = image.getHeight()*showLines/level.getLines();
				if (height>0){
					BufferedImage img = image.getSubimage(0, 0, width, height);
					g.drawImage(img, x, y, null);
				}
			}
		} else if (isStateNextLevelInfo(currentState)){
			g.setColor(Color.BLACK);
			g.fillRect(getBounds().x, getBounds().y, getBounds().width, getBounds().height);
			g.setColor(Color.RED);
			Graphics2D g2D = (Graphics2D) g;
			Font font = g.getFont();
			font = font.deriveFont(36);
			font = font.deriveFont(Font.BOLD);
			int x = getBounds().x+TetrisConstants.PICTURE_VIEW_BORDER;
			int y = getBounds().y+getBounds().height/3;
			int h;
			GlyphVector gv = createGlyphVector(g2D, "Level .....: "+nextLevel.getLevel(), font);
			h = gv.getVisualBounds().getBounds().height;
			g2D.drawGlyphVector(gv, x, y);
			gv = createGlyphVector(g2D, "Lines ....: "+nextLevel.getLines(), font);
			y += 2*h;
			h = gv.getVisualBounds().getBounds().height;
			g2D.drawGlyphVector(gv, x, y);
			gv = createGlyphVector(g2D, "Image ....: ", font);
			y += 2*h;
			h = gv.getVisualBounds().getBounds().height;
			g2D.drawGlyphVector(gv, x, y);
			String fileName = fileNames[computeImageIndex()];
			fileName = fileName.substring(0, fileName.lastIndexOf("."));
			fileName = WordUtils.wrap(fileName, 80);
			String[] parts = fileName.split(System.getProperty("line.separator"));
			for (String line : parts) {
				gv = createGlyphVector(g2D, line, font);
				y += 2*h;
				g2D.drawGlyphVector(gv, x, y);
			}
		}
		g.setColor(oldColor);
	}
	/**
	 * Compute the index of the image/filename corresponding to the current level.
	 * @return
	 */
	private int computeImageIndex() {
		int index = (level.getLevel()-1)%images.length;
		return index;
	}

	/**
	 * Handle the update events
	 */
	public void gameUpdate(Event e, Object notice) {
		if (e instanceof RowCompleteEvent){
			showLines++;
		}
		if (e instanceof NextLevelEvent){
			nextLevel = (Level) notice;
			if (nextLevel.getLevel() == 1){
				level = nextLevel;
			}
		}
		if (e instanceof GameStateChangeEvent) {
			GameStateChangeEvent change = (GameStateChangeEvent) e;
			currentState = change.getNextGameState();
			if (change.getNextGameState() == EGameState.NEXT_LEVEL_INFO){
				showLines = 0;
				level = nextLevel;
			}
		}
		if (e instanceof KeyStrokeEvent){
			if (notice == EKeyStroke.PAUSE && (isPlayState(currentState) || isPauseState(currentState))) {
				currentState = togglePause(currentState);
			}
		}
	}
	/**
	 * Filter for reading image files
	 * @author andi
	 *
	 */
	private static class FilenameFilter implements java.io.FilenameFilter {
		
		private final String[] acceptedExtensions = new String[]{"jpg","jpeg","png"};

		public boolean accept(File dir, String name) {
			if (name.contains(".")){
				String lName = name.toLowerCase();
				for (String extension : acceptedExtensions) {
					if (lName.endsWith(extension)) {
						return true;
					}
				}
			}
			return false;
		}
	}

}
