package ch.sahits.game.tetris.ui;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.font.GlyphVector;

import ch.sahits.game.event.Event;
import ch.sahits.game.rendering.AbstractRenderPart;
import ch.sahits.game.tetris.ctrl.GameStateUtility;
import ch.sahits.game.tetris.event.EGameState;
import ch.sahits.game.tetris.event.GameStateChangeEvent;
import ch.sahits.game.tetris.event.NextShapeEvent;
import ch.sahits.game.tetris.event.PlayFieldUpdateEvent;
import ch.sahits.game.tetris.model.Block;
import ch.sahits.game.tetris.model.EBuildingBlocks;
import ch.sahits.game.tetris.model.PlayingField;
import ch.sahits.game.tetris.model.TetrisViewModel;
import static ch.sahits.game.tetris.ctrl.GameStateUtility.*;
import static ch.sahits.game.tetris.ui.RenderingUtilities.*;
/**
 * The classic tetris view
 * @author andi
 *
 */
public class TetrisView extends AbstractRenderPart {

	private EBuildingBlocks nextShape;
	private TetrisViewModel model;
	private EGameState currentState = EGameState.START;

	public TetrisView(Rectangle rect) {
		super(rect);
		Event.add(this);
	}

	public void gameRender(Graphics g) {
		Color origColor = g.getColor();
		Rectangle rect = getBounds();
		g.setColor(Color.BLACK);
		g.fillRect(rect.x, rect.y, rect.width, rect.height);
		int topLeftX = rect.x;
		int topLeftY = rect.y-30;
		int width = TetrisConstants.BLOCK_SIZE*TetrisConstants.BLOCKS_PER_ROW+2*TetrisConstants.BORDER_THICKNESS;
		Rectangle border = new Rectangle(topLeftX, topLeftY, width, TetrisConstants.BLOCK_SIZE*TetrisConstants.NUMBER_OF_ROWS+TetrisConstants.BORDER_THICKNESS);
		Rectangle inner = new Rectangle(border.x+TetrisConstants.BORDER_THICKNESS,border.y,TetrisConstants.BLOCK_SIZE*TetrisConstants.BLOCKS_PER_ROW,border.height-TetrisConstants.BORDER_THICKNESS);
		renderGamePlay(g, border, inner);
		renderMenu(g, border, inner);
		g.setColor(origColor);
	}
	/**
	 * Rendering when the menu is shown
	 * @param g
	 * @param border
	 * @param inner
	 */
	private void renderMenu(Graphics g, Rectangle border, Rectangle inner) {
		if (isMenuState(currentState) || isKeySetState(currentState)){
			drawBlueBox(g, border, inner);
		}
		
	}
	/**
	 * Rendering during game play
	 * @param g
	 * @param border
	 * @param inner
	 */
	private void renderGamePlay(Graphics g, Rectangle border, Rectangle inner) {
		if (isGamePlayState(currentState)) {
			drawBlueBox(g, border, inner);
			drawPlayingField(g,inner.x,inner.y);
			if (!GameStateUtility.isStateNextLevelInfo(currentState)){ 
				drawNextShape(g,border);
			}
		}
		if (isGameOverState(currentState)) {
			drawGameOver(g,inner);
		}
	}
	/**
	 * Draw the blue tetris border
	 * @param g
	 * @param border
	 * @param inner
	 */
	private void drawBlueBox(Graphics g, Rectangle border, Rectangle inner) {
		g.setColor(Color.BLUE);
		g.fillRect(border.x, border.y, border.width, border.height);
		g.setColor(Color.BLACK);
		g.fillRect(inner.x, inner.y, inner.width, inner.height);
	}
	/**
	 * Draw the game over message
	 * @param g
	 * @param inner
	 */
	private void drawGameOver(Graphics g, Rectangle inner) {
		int x = inner.x;
		int y = inner.y+TetrisConstants.BLOCK_SIZE*8;
		g.fillRect(x, y, inner.width, TetrisConstants.BLOCK_SIZE);
		Font font = g.getFont();
		font = font.deriveFont(18);
		Color oldColor = g.getColor();
		g.setColor(Color.RED);
		GlyphVector gv = createGlyphVector((Graphics2D) g, "Game Over", font);
		Rectangle messageBox = gv.getVisualBounds().getBounds();
		int messageX = x+(inner.width-messageBox.width)/2;
		((Graphics2D) g).drawGlyphVector(gv, messageX, y+TetrisConstants.BLOCK_SIZE-5);
		g.setColor(oldColor);
	}
	/**
	 * Draw the playing field
	 * @param g
	 * @param x
	 * @param y
	 */
	private void drawPlayingField(Graphics g,int x, int y) {
		Block block = model.getBlock();
		Point[] positions = block.getPosition();
		for (Point point : positions) {
			int topLeftX = x+(point.x*TetrisConstants.BLOCK_SIZE);
			int topLeftY = y+(point.y*TetrisConstants.BLOCK_SIZE);
			drawBlock(g, block.getColor(), topLeftX, topLeftY);
		}
		PlayingField field = model.getPlayingField();
		for (int yIndex=0;yIndex<field.getNbRows();yIndex++){
			for (int xIndex=0;xIndex<field.getNbColumns();xIndex++){
				Color c = field.get(xIndex, yIndex); 
				if (c!=null){
					drawBlock(g, c, x+xIndex*TetrisConstants.BLOCK_SIZE, y+yIndex*TetrisConstants.BLOCK_SIZE);
				}
			}
		}
	}
	/**
	 * Draw the next shape
	 * @param g
	 * @param tetrisBoard
	 */
	private void drawNextShape(Graphics g,Rectangle tetrisBoard) {
		int x = tetrisBoard.x+tetrisBoard.width/2-TetrisConstants.BLOCK_SIZE/2;
		int y = tetrisBoard.y+tetrisBoard.height+TetrisConstants.MARGIN;
		switch (nextShape) {
		case L_LEFT: {
			x = tetrisBoard.x+tetrisBoard.width/2-TetrisConstants.BLOCK_SIZE;
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE*2);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE*2);
			break;
		}
		case L_RIGHT: {
			x = tetrisBoard.x+tetrisBoard.width/2-TetrisConstants.BLOCK_SIZE;
			drawBlock(g, nextShape.getColor(), x, y);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE*2);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE*2);
			break;
		}
		case CUBE: {
			x = tetrisBoard.x+tetrisBoard.width/2-TetrisConstants.BLOCK_SIZE;
			drawBlock(g, nextShape.getColor(), x, y);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE);
			break;
		}
		case LEFT: {
			x = tetrisBoard.x+tetrisBoard.width/2-TetrisConstants.BLOCK_SIZE;
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE*2);
			break;
		}
		case RIGHT: {
			x = tetrisBoard.x+tetrisBoard.width/2-TetrisConstants.BLOCK_SIZE;
			drawBlock(g, nextShape.getColor(), x, y);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE*2);
			break;
		}
		case I: {
			drawBlock(g, nextShape.getColor(), x, y);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE*2);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE*3);
			break;
		}
		case T: {
			x = x-TetrisConstants.BLOCK_SIZE;
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y);
			drawBlock(g, nextShape.getColor(), x, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE, y+TetrisConstants.BLOCK_SIZE);
			drawBlock(g, nextShape.getColor(), x+TetrisConstants.BLOCK_SIZE*2, y+TetrisConstants.BLOCK_SIZE);
			break;

		}
		default:
		{
			throw new IllegalStateException("Must have nextShape: "+nextShape);
		}
		} // end switch
		
	}
	/**
	 * Draw a plock
	 * @param g
	 * @param c Color
	 * @param x top left x
	 * @param y top left y
	 */
	private void drawBlock(Graphics g, Color c, int x, int y) {
		Color oldColor = g.getColor();
		g.setColor(c);
		g.fill3DRect(x, y, TetrisConstants.BLOCK_SIZE, TetrisConstants.BLOCK_SIZE, true);
		g.setColor(oldColor);
	}
	/**
	 * Keep track of game state, playing field and next shape updates.
	 */
	public void gameUpdate(Event e, Object notice) {
		if (e instanceof PlayFieldUpdateEvent){
			model = ((PlayFieldUpdateEvent)e).getModel();
		}
		if (e instanceof NextShapeEvent) {
			nextShape = (EBuildingBlocks) notice;
		}
		if (e instanceof GameStateChangeEvent) {
			GameStateChangeEvent change = (GameStateChangeEvent) e;
			currentState = change.getNextGameState();
		}
	}
	


}
